""" Active Learning by QUerying Informative and Representative Examples (QUIRE)

This module contains a class that implements an active learning algorithm
(query strategy): QUIRE

"""

import bisect

import numpy as np
from sklearn.metrics.pairwise import rbf_kernel

from libact.base.interfaces import QueryStrategy


class QUIRE(QueryStrategy):
    """Querying Informative and Representative Examples (QUIRE)

    Query the most informative and representative examples where the metrics
    measuring and combining are done using min-max approach.

    Parameters
    ----------
    lmbda: float, optional (default=1.0)
        A regularization parameter used in the regularization learning framework.

    gamma: float, optional (default=1.0)
        A parameter for computing rbf kernel.

    K: sklearn.metrics.pairwise.*_kernel, optional (default = 'rbf')
        Kernel matrix. Currently supports only rbf kernel.

    Attributes
    ----------

    Reference
    ---------

    .. [1] S.-J. Huang, R. Jin, and Z.-H. Zhou. Active learning by querying
           informative and representative examples.
    """

    def __init__(self, *args, **kwargs):
        super(QUIRE, self).__init__(*args, **kwargs)
        self.Uindex = [
            idx for idx, _ in self.dataset.get_unlabeled_entries()
            ]
        self.Lindex = [
            idx for idx in range(len(self.dataset)) if idx not in self.Uindex
            ]
        self.lmbda = kwargs.pop('lmbda', 1.)
        self.gamma = kwargs.pop('gamma', 1.)
        X, self.y = zip(*self.dataset.get_entries())
        self.y = list(self.y)
        K = rbf_kernel(X=X, Y=X, gamma=self.gamma) 
        #TODO: extend for other kernel functions
        self.K = K
        self.L = np.linalg.inv(K + self.lmbda * np.eye(len(X)))

    def update(self, entry_id, label):
        bisect.insort(a=self.Lindex, x=entry_id)
        self.Uindex.remove(entry_id)
        self.y[entry_id] = label

    def make_query(self):
        L = self.L
        Lindex = self.Lindex
        Uindex = self.Uindex
        query_index = -1
        min_eva = np.inf
        y_labeled = np.array([label for label in self.y if label is not None])
        det_Laa = np.linalg.det(L[np.ix_(Uindex, Uindex)])
        # efficient computation of inv(Laa)
        M3 = np.dot(self.K[np.ix_(Uindex, Lindex)],
                    np.linalg.inv(self.lmbda * np.eye(len(Lindex))))
        M2 = np.dot(M3, self.K[np.ix_(Lindex, Uindex)])
        M1 = self.lmbda * np.eye(len(Uindex)) + self.K[np.ix_(Uindex, Uindex)]
        inv_Laa = M1 - M2
        iList = list(range(len(Uindex)))
        if len(iList) == 1:
            return Uindex[0]
        for i, each_index in enumerate(Uindex):
            # go through all unlabeled instances and compute their evaluation
            # values one by one
            Uindex_r = Uindex[:]
            Uindex_r.remove(each_index)
            iList_r = iList[:]
            iList_r.remove(i)
            inv_Luu = inv_Laa[np.ix_(iList_r, iList_r)] - 1/inv_Laa[i, i] * \
                np.dot(inv_Laa[iList_r, i], inv_Laa[iList_r, i].T)
            tmp = np.dot(
                      L[each_index][Lindex] - \
                          np.dot(
                              np.dot(
                                  L[each_index][Uindex_r], 
                                  inv_Luu
                              ),
                              L[np.ix_(Uindex_r, Lindex)]
                          ),
                      y_labeled,
                  )
            eva = L[each_index][each_index] - \
                det_Laa / L[each_index][each_index] + 2 * np.abs(tmp)

            if eva < min_eva:
                query_index = each_index
                min_eva = eva
        return query_index
